// Copyright (C) 2020-2021 Joel Rosdahl and other contributors
//
// See doc/AUTHORS.adoc for a complete list of contributors.
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 3 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 51
// Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

#pragma once

#include <cstdint>
#include <string>

class Lockfile
{
public:
  // Acquire a lock on `path`. Break the lock (or give up, depending on
  // implementation) after `staleness_limit` Microseconds.
  Lockfile(const std::string& path, uint32_t staleness_limit = 2000000);

  // Release the lock if acquired.
  ~Lockfile();

  // Return whether the lockfile was acquired successfully.
  bool acquired() const;

private:
  std::string m_lockfile;
#ifndef _WIN32
  bool m_acquired = false;
#else
  void* m_handle = nullptr;
#endif
};
